# MIT License
#
# Copyright (c) 2018 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time

from requests.exceptions import Timeout as RequestsTimeout, HTTPError, TooManyRedirects

from query_state_lib.base.executors.bounded_executor import BoundedExecutor
from query_state_lib.base.executors.fail_safe_executor import FailSafeExecutor
from query_state_lib.base.executors.utils.progress_logger import ProgressLogger
from query_state_lib.base.executors.utils.retriable_value_error import RetriableValueError
from query_state_lib.base.executors.utils.utils import dynamic_batch_iterator

RETRY_EXCEPTIONS = (ConnectionError, HTTPError, RequestsTimeout, TooManyRedirects, OSError,
                    RetriableValueError)

BATCH_CHANGE_COOLDOWN_PERIOD_SECONDS = 2 * 60


# Executes the given work in batches, reducing the batch size exponentially in case of errors.
class BatchWorkExecutor:
    def __init__(self, starting_batch_size, max_workers, retry_exceptions=RETRY_EXCEPTIONS, max_retries=5, sleep=10):
        self.batch_size = starting_batch_size
        self.max_batch_size = starting_batch_size
        self.latest_batch_size_change_time = None
        self.max_workers = max_workers
        # Using bounded executor prevents unlimited queue growth
        # and allows monitoring in-progress futures and failing fast in case of errors.
        self.executor = FailSafeExecutor(BoundedExecutor(1, self.max_workers))
        self.retry_exceptions = retry_exceptions
        self.max_retries = max_retries
        self.progress_logger = ProgressLogger()
        self.sleep = sleep
        self.logger = logging.getLogger('BatchWorkExecutor')

    def execute(self, work_iterable, work_handler, total_items=None):
        self.progress_logger.start(total_items=total_items)
        for batch in dynamic_batch_iterator(work_iterable, lambda: self.batch_size):
            self.executor.submit(self._fail_safe_execute, work_handler, batch)

    def _fail_safe_execute(self, work_handler, batch):
        try:
            work_handler(batch)
            self._try_increase_batch_size(len(batch))
        except self.retry_exceptions:
            self.logger.exception('An exception occurred while executing work_handler.')
            self._try_decrease_batch_size(len(batch))
            self.logger.info('The batch of size {} will be retried one item at a time.'.format(len(batch)))
            time.sleep(self.sleep)
            execute_with_retries(work_handler, batch,
                                 max_retries=self.max_retries, retry_exceptions=self.retry_exceptions)

        self.progress_logger.track(len(batch))

    # Some acceptable race conditions are possible
    def _try_decrease_batch_size(self, current_batch_size):
        batch_size = self.batch_size
        if batch_size == current_batch_size and batch_size > 1:
            new_batch_size = int(current_batch_size / 2)
            self.logger.info('Reducing batch size to {}.'.format(new_batch_size))
            self.batch_size = new_batch_size
            self.latest_batch_size_change_time = time.time()

    def _try_increase_batch_size(self, current_batch_size):
        if current_batch_size * 2 <= self.max_batch_size:
            current_time = time.time()
            latest_batch_size_change_time = self.latest_batch_size_change_time
            seconds_since_last_change = current_time - latest_batch_size_change_time \
                if latest_batch_size_change_time is not None else 0
            if seconds_since_last_change > BATCH_CHANGE_COOLDOWN_PERIOD_SECONDS:
                new_batch_size = current_batch_size * 2
                self.logger.info('Increasing batch size to {}.'.format(new_batch_size))
                self.batch_size = new_batch_size
                self.latest_batch_size_change_time = current_time

    def shutdown(self):
        self.executor.shutdown()
        self.progress_logger.finish()


def execute_with_retries(func, *args, max_retries=5, retry_exceptions=RETRY_EXCEPTIONS, sleep_seconds=1):
    for i in range(max_retries):
        try:
            return func(*args)
        except retry_exceptions:
            logging.exception('An exception occurred while executing execute_with_retries. Retry #{}'.format(i))
            if i < max_retries - 1:
                logging.info('The request will be retried after {} seconds. Retry #{}'.format(sleep_seconds, i))
                time.sleep(sleep_seconds)
                continue
            else:
                raise



import random
from abc import ABC, abstractmethod


class HostResolver(ABC):  # pragma: no cover
    """Abstract base class for host resolvers."""

    @abstractmethod
    def get_host_index(self) -> int:
        raise NotImplementedError


class SingleHostResolver(HostResolver):
    """Single host resolver."""

    def __init__(self, x_api_key):
        self.x_api_key = x_api_key

    def get_host_index(self) -> int:
        return 0


class RandomHostResolver(HostResolver):
    """Random host resolver."""

    def __init__(self, host_count: int) -> None:
        self._max = host_count - 1

    def get_host_index(self) -> int:
        return random.randint(0, self._max)


class RoundRobinHostResolver(HostResolver):
    """Round-robin host resolver."""

    def __init__(self, host_count: int) -> None:
        self._index = -1
        self._count = host_count

    def get_host_index(self) -> int:
        self._index = (self._index + 1) % self._count
        return self._index



# MIT License
#
# Copyright (c) 2018 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import itertools
import warnings

from query_state_lib.base.executors.utils.retriable_value_error import RetriableValueError


def hex_to_dec(hex_string):
    if hex_string is None:
        return None
    try:
        return int(hex_string, 16)
    except ValueError:
        print("Not a hex string %s" % hex_string)
        return hex_string


def to_int_or_none(val):
    if isinstance(val, int):
        return val
    if val is None or val == '':
        return None
    try:
        return int(val)
    except ValueError:
        return None


def chunk_string(string, length):
    return (string[0 + i:length + i] for i in range(0, len(string), length))


def to_normalized_address(address):
    if address is None or not isinstance(address, str):
        return address
    return address.lower()


def validate_range(range_start_incl, range_end_incl):
    if range_start_incl < 0 or range_end_incl < 0:
        raise ValueError('range_start and range_end must be greater or equal to 0')

    if range_end_incl < range_start_incl:
        raise ValueError('range_end must be greater or equal to range_start')


def rpc_response_batch_to_results(response):
    for response_item in response:
        yield rpc_response_to_result(response_item)


def rpc_response_to_result(response):
    result = response.get('result')
    if result is None:
        error_message = 'result is None in response {}.'.format(response)
        if response.get('error') is None:
            error_message = error_message + ' Make sure Ethereum node is synced.'
            # When nodes are behind a load balancer it makes sense to retry the request in hopes it will go to other,
            # synced node
            raise RetriableValueError(error_message)
        elif response.get('error') is not None and is_retriable_error(response.get('error').get('code')):
            raise RetriableValueError(error_message)
        raise ValueError(error_message)
    return result


def is_retriable_error(error_code):
    if error_code is None:
        return False

    if not isinstance(error_code, int):
        return False

    # https://www.jsonrpc.org/specification#error_object
    if error_code == -32603 or (-32000 >= error_code >= -32099):
        return True

    return False


def split_to_batches(start_incl, end_incl, batch_size):
    """start_incl and end_incl are inclusive, the returned batch ranges are also inclusive"""
    for batch_start in range(start_incl, end_incl + 1, batch_size):
        batch_end = min(batch_start + batch_size - 1, end_incl)
        yield batch_start, batch_end


def dynamic_batch_iterator(iterable, batch_size_getter):
    batch = []
    batch_size = batch_size_getter()
    for item in iterable:
        batch.append(item)
        if len(batch) >= batch_size:
            yield batch
            batch = []
            batch_size = batch_size_getter()
    if len(batch) > 0:
        yield batch


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def check_classic_provider_uri(chain, provider_uri):
    if chain == 'classic' and provider_uri == 'https://mainnet.infura.io':
        warnings.warn("ETC Chain not supported on Infura.io. Using https://ethereumclassic.network instead")
        return 'https://ethereumclassic.network'
    return provider_uri



import logging

from query_state_lib.jobs.handler_batch_request_job import HandlerBatchRequestJob


def get_batch_request(json_request):
    return json_request.get("request")


def get_batch_response(json_response):
    return {"response": json_response}


def divide_chunks(l, n):
    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]


def is_batch_request(request):
    try:
        return request.get("batch_request")
    except Exception as e:
        return False


class StateQuerierServer:
    def __init__(self, batch_size=2000, max_workers=8, logger=logging.getLogger("StateQuerierServer")):
        """
        process_request is func with signature  same as : def process(request, batch=True)
        """
        self.batch_size = batch_size
        self.max_workers = max_workers
        self.logger = logger

    def generate_batch_query_request(self, json_request):
        batch_request = get_batch_request(json_request)
        return list(divide_chunks(batch_request, self.batch_size))

    def handle_batch_request(self, request_json, process_func):
        try:
            request_data = get_batch_request(request_json)
            job = HandlerBatchRequestJob(request_data, process_func, self.batch_size, self.max_workers)
            job.run()
            response = job.get_response()
            return {"response": response}
        except Exception as e:
            self.logger.warning(f"err when handle batch request {e}")
            return None



from typing import Any, MutableMapping, Optional


class Response:
    """HTTP response.

    :param method: HTTP method in lowercase (e.g. "post").
    :type method: str
    :param url: API URL.
    :type url: str
    :param headers: Response headers.
    :type headers: MutableMapping
    :param status_code: Response status code.
    :type status_code: int
    :param status_text: Response status text.
    :type status_text: str
    :param raw_body: Raw response body.
    :type raw_body: str

    :ivar method: HTTP method in lowercase (e.g. "post").
    :vartype method: str
    :ivar url: API URL.
    :vartype url: str
    :ivar headers: Response headers.
    :vartype headers: MutableMapping
    :ivar status_code: Response status code.
    :vartype status_code: int
    :ivar status_text: Response status text.
    :vartype status_text: str
    :ivar raw_body: Raw response body.
    :vartype raw_body: str
    :ivar body: JSON-deserialized response body.
    :vartype body: str | bool | int | float | list | dict | None
    :ivar error_code: Error code from ArangoDB server.
    :vartype error_code: int
    :ivar error_message: Error message from ArangoDB server.
    :vartype error_message: str
    :ivar is_success: True if response status code was 2XX.
    :vartype is_success: bool
    """

    __slots__ = (
        "method",
        "url",
        "headers",
        "status_code",
        "status_text",
        "body",
        "raw_body",
        "error_code",
        "error_message",
        "is_success",
    )

    def __init__(
            self,
            method: str,
            url: str,
            headers: MutableMapping[str, str],
            status_code: int,
            status_text: str,
            raw_body: str,
    ) -> None:
        self.method = method.lower()
        self.url = url
        self.headers = headers
        self.status_code = status_code
        self.status_text = status_text
        self.raw_body = raw_body

        # Populated later
        self.body: Any = None
        self.error_code: Optional[int] = None
        self.error_message: Optional[str] = None
        self.is_success: Optional[bool] = None



from abc import ABC, abstractmethod
from typing import MutableMapping, Optional, Tuple, Union

from requests import Session
from requests.adapters import HTTPAdapter
from requests_toolbelt import MultipartEncoder
from urllib3.util.retry import Retry

from query_state_lib.client.moralis_client.moralis_response import Response
from query_state_lib.client.moralis_client.typings import Headers


class HTTPClient(ABC):  # pragma: no cover
    """Abstract base class for HTTP clients."""

    @abstractmethod
    def create_session(self, host: str) -> Session:
        """Return a new requests session given the host URL.

        This method must be overridden by the user.

        :param host: Moralis host URL.
        :type host: str
        :returns: Requests session object.
        :rtype: requests.Session
        """
        raise NotImplementedError

    @abstractmethod
    def send_request(
            self,
            session: Session,
            method: str,
            url: str,
            headers: Optional[Headers] = None,
            params: Optional[MutableMapping[str, str]] = None,
            data: Union[str, MultipartEncoder, None] = None,
            auth: Optional[Tuple[str, str]] = None,
    ) -> Response:
        """Send an HTTP request.

        This method must be overridden by the user.

        :param session: Requests session object.
        :type session: requests.Session
        :param method: HTTP method in lowercase (e.g. "post").
        :type method: str
        :param url: Request URL.
        :type url: str
        :param headers: Request headers.
        :type headers: dict
        :param params: URL (query) parameters.
        :type params: dict
        :param data: Request payload.
        :type data: str | MultipartEncoder | None
        :param auth: Username and password.
        :type auth: tuple
        :returns: HTTP response.
        :rtype: arango.response.Response
        """
        raise NotImplementedError


class DefaultHTTPClient(HTTPClient):
    """Default HTTP client implementation."""

    REQUEST_TIMEOUT = 60
    RETRY_ATTEMPTS = 3
    BACKOFF_FACTOR = 1

    def create_session(self, host: str) -> Session:
        """Create and return a new session/connection.

        :param host: Moralis host URL.
        :type host: str
        :returns: requests session object
        :rtype: requests.Session
        """
        retry_strategy = Retry(
            total=self.RETRY_ATTEMPTS,
            backoff_factor=self.BACKOFF_FACTOR,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "OPTIONS"],
        )
        http_adapter = HTTPAdapter(max_retries=retry_strategy)

        session = Session()
        session.mount("https://", http_adapter)
        session.mount("http://", http_adapter)

        return session

    def send_request(
            self,
            session: Session,
            method: str,
            url: str,
            headers: Optional[Headers] = None,
            params: Optional[MutableMapping[str, str]] = None,
            data: Union[str, MultipartEncoder, None] = None,
            auth: Optional[Tuple[str, str]] = None,
    ) -> Response:
        """Send an HTTP request.

        :param session: Requests session object.
        :type session: requests.Session
        :param method: HTTP method in lowercase (e.g. "post").
        :type method: str
        :param url: Request URL.
        :type url: str
        :param headers: Request headers.
        :type headers: dict
        :param params: URL (query) parameters.
        :type params: dict
        :param data: Request payload.
        :type data: str | MultipartEncoder | None
        :param auth: Username and password.
        :type auth: tuple
        :returns: HTTP response.
        :rtype: arango.response.Response
        """
        response = session.request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=self.REQUEST_TIMEOUT,
        )
        return Response(
            method=method,
            url=response.url,
            headers=response.headers,
            status_code=response.status_code,
            status_text=response.reason,
            raw_body=response.text,
        )



from json import dumps, loads
from typing import Optional, Sequence, Callable, Any

from query_state_lib.client.moralis_client.connection import BasicConnection
from query_state_lib.client.moralis_client.examples import LIQUIDATE_EVENT
from query_state_lib.client.moralis_client.moralis_http import HTTPClient, DefaultHTTPClient
from query_state_lib.client.moralis_client.moralis_request import Request
from query_state_lib.client.moralis_client.resolver import HostResolver, SingleHostResolver, RandomHostResolver, \
    RoundRobinHostResolver


class MoralisClient:
    def __init__(self, x_api_keys, base_url="https://deep-index.moralis.io/api/v2",
                 host_resolver: str = "roundrobin",
                 http_client: Optional[HTTPClient] = None,
                 serializer: Callable[..., str] = lambda x: dumps(x),
                 deserializer: Callable[[str], Any] = lambda x: loads(x)):
        self.base_url = base_url

        if isinstance(x_api_keys, str):
            self._x_api_keys = [host.strip("/") for host in x_api_keys.split(",")]
        else:
            self._x_api_keys = [host.strip("/") for host in x_api_keys]

        num_x_api_key = len(self._x_api_keys)
        self._host_resolver: HostResolver

        if num_x_api_key == 1:
            self._host_resolver = SingleHostResolver(x_api_key=num_x_api_key)
        elif host_resolver == "random":
            self._host_resolver = RandomHostResolver(num_x_api_key)
        else:
            self._host_resolver = RoundRobinHostResolver(num_x_api_key)
        self._http = http_client or DefaultHTTPClient()
        self._serializer = serializer
        self._deserializer = deserializer
        self._sessions = [self._http.create_session(h) for h in self._x_api_keys]

        self.connection = BasicConnection(
            host=self.base_url,
            x_api_keys=x_api_keys,
            host_resolver=self._host_resolver,
            sessions=self._sessions,
            http_client=self._http,
            serializer=self._serializer,
            deserializer=self._deserializer,
        )

    def __repr__(self) -> str:
        return f"<ArangoClient {','.join(self._x_api_keys)}>"

    def close(self) -> None:  # pragma: no cover
        """Close HTTP sessions."""
        for session in self._sessions:
            session.close()

    @property
    def x_api_keys(self) -> Sequence[str]:
        """Return the list of ArangoDB host URLs.

        :return: List of ArangoDB host URLs.
        :rtype: [str]
        """
        return self._x_api_keys

    def ping(self):
        return self.connection.ping()

    def date_to_block(self, chain="eth", date=None):
        endpoint = f"/dateToBlock"
        params = {
            "chain": chain
        }
        if date:
            params["date"] = date

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_token_price(self, address, chain="eth", to_block=None, exchange=None):
        endpoint = f"/erc20/{address}/price"
        params = {
            "chain": chain
        }
        if to_block:
            params["to_block"] = to_block
        if exchange:
            params["exchange"] = exchange

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_token_metadata(self, addresses, chain="eth"):
        endpoint = f"/erc20/metadata"
        params = {
            "chain": chain
        }
        if addresses:
            params["addresses"] = addresses

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_token_balance(self, address, chain="eth", to_block=None, token_addresses=[]):
        endpoint = f"/{address}/erc20"
        params = {
            "chain": chain
        }
        if to_block:
            params["to_block"] = to_block
        if token_addresses:
            params["token_addresses"] = token_addresses

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_address_transfers(self, address,
                              chain="eth",
                              from_block=None,
                              to_block=None,
                              from_date=None,
                              to_date=None,
                              offset=None,
                              limit=None,
                              ):
        endpoint = f"/{address}/erc20/transfers"
        params = {
            "chain": chain
        }
        if from_block:
            params["from_block"] = from_block
        if to_block:
            params["to_block"] = to_block
        if from_date:
            params["from_date"] = from_date
        if to_date:
            params["to_date"] = to_date
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_address_logs(self, address, chain="eth",
                         block_number=None,
                         from_block=None,
                         to_block=None,
                         from_date=None,
                         to_date=None,
                         topic0=None,
                         topic1=None,
                         topic2=None,
                         topic3=None,
                         offset=None,
                         limit=None,
                         ):
        endpoint = f"/{address}/logs"
        params = {
            "chain": chain
        }
        if block_number:
            params["block_number"] = block_number
        if from_block:
            params["from_block"] = from_block
        if to_block:
            params["to_block"] = to_block
        if from_date:
            params["from_date"] = from_date
        if to_date:
            params["to_date"] = to_date
        if topic0:
            params["topic0"] = topic0
        if topic1:
            params["topic1"] = topic1
        if topic2:
            params["topic2"] = topic2
        if topic3:
            params["topic3"] = topic3
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit

        request = Request(method="get", endpoint=endpoint, params=params)
        res = self.connection.send_request(request)
        return res.body

    def get_address_event_by_topic(self, address,
                                   topic,
                                   event_abi,
                                   chain="eth",
                                   block_number=None,
                                   from_block=None,
                                   to_block=None,
                                   from_date=None,
                                   to_date=None,
                                   offset=None,
                                   limit=None,
                                   ):
        endpoint = f"/{address}/events"
        params = {
            "chain": chain,
            "topic": topic
        }
        if block_number:
            params["block_number"] = block_number
        if from_block:
            params["from_block"] = from_block
        if to_block:
            params["to_block"] = to_block
        if from_date:
            params["from_date"] = from_date
        if to_date:
            params["to_date"] = to_date
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit

        request = Request(method="post", endpoint=endpoint, params=params, data=event_abi)
        res = self.connection.send_request(request)
        return res.body



from web3 import Web3

w3 = Web3()


def hex_to_dec(hex_string):
    if hex_string is None:
        return None
    try:
        return int(hex_string, 16)
    except ValueError:
        print("Not a hex string %s" % hex_string)
        return hex_string


def decode_output_rpc(types, result):
    bytes_data = bytes.fromhex(result[2:])
    decode_data = w3.codec.decode_abi(types, bytes_data)
    return decode_data


def decode_eth_call_data(abi, fn_name, result):
    fn_abi = None
    for _fn_abi in abi:
        if _fn_abi.get("name") == fn_name:
            fn_abi = _fn_abi
            break
    if not fn_abi:
        raise Exception(f"function {fn_name} not found in abi")
    outputs = fn_abi.get("outputs")
    types = []
    for output in outputs:
        if output.get("type") == "tuple":
            _tuple_type = _append_output_tuple_type(output)
            types.append(_tuple_type)
        else:
            types.append(output.get("type"))

    return decode_output_rpc(types, result)


def _append_output_tuple_type(output):
    components = output.get("components")
    _tuple_type = '('
    for component in components:
        _type_str = component.get('type')
        if _type_str == "tuple":
            _tuple_type += _append_output_tuple_type(component) + ","
        else:
            _tuple_type += _type_str + ","
    _tuple_type = _tuple_type[0:-1]+")"
    return _tuple_type


def decode_eth_call_balance_of(result):
    types = ["uint256"]
    return decode_output_rpc(types, result)[0]


def decode_eth_get_balance(result):
    return hex_to_dec(result)



import sys
import time
from abc import abstractmethod
from typing import Any, Callable, Optional, Sequence, Union

import jwt
from requests import Session
from requests_toolbelt import MultipartEncoder

from query_state_lib.client.moralis_client.exceptions import ServerConnectionError
from query_state_lib.client.moralis_client.moralis_http import HTTPClient
from query_state_lib.client.moralis_client.moralis_request import Request
from query_state_lib.client.moralis_client.resolver import HostResolver
from query_state_lib.client.moralis_client.moralis_response import Response
from query_state_lib.client.moralis_client.typings import Fields, Json


class BaseConnection:
    """Base connection to a specific ArangoDB database."""

    def __init__(
            self,
            host: str,
            _x_api_keys: Fields,
            host_resolver: HostResolver,
            sessions: Sequence[Session],
            http_client: HTTPClient,
            serializer: Callable[..., str],
            deserializer: Callable[[str], Any],
    ) -> None:
        self._url_prefix = host
        self._x_api_keys = _x_api_keys
        self._host_resolver = host_resolver
        self._sessions = sessions
        self._http = http_client
        self._serializer = serializer
        self._deserializer = deserializer

    def serialize(self, obj: Any) -> str:
        """Serialize the given object.

        :param obj: JSON object to serialize.
        :type obj: str | bool | int | float | list | dict | None
        :return: Serialized string.
        :rtype: str
        """
        return self._serializer(obj)

    def deserialize(self, string: str) -> Any:
        """De-serialize the string and return the object.

        :param string: String to de-serialize.
        :type string: str
        :return: De-serialized JSON object.
        :rtype: str | bool | int | float | list | dict | None
        """
        try:
            return self._deserializer(string)
        except (ValueError, TypeError):
            return string

    def prep_response(self, resp: Response, deserialize: bool = True) -> Response:
        """Populate the response with details and return it.

        :param deserialize: Deserialize the response body.
        :type deserialize: bool
        :param resp: HTTP response.
        :type resp: arango.response.Response
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        if deserialize:
            resp.body = self.deserialize(resp.raw_body)
            if isinstance(resp.body, dict):
                resp.error_code = resp.body.get("errorNum")
                resp.error_message = resp.body.get("errorMessage")
        else:
            resp.body = resp.raw_body

        http_ok = 200 <= resp.status_code < 300
        resp.is_success = http_ok and resp.error_code is None
        return resp

    def prep_bulk_err_response(self, parent_response: Response, body: Json) -> Response:
        """Build and return a bulk error response.

        :param parent_response: Parent response.
        :type parent_response: arango.response.Response
        :param body: Error response body.
        :type body: dict
        :return: Child bulk error response.
        :rtype: arango.response.Response
        """
        resp = Response(
            method=parent_response.method,
            url=parent_response.url,
            headers=parent_response.headers,
            status_code=parent_response.status_code,
            status_text=parent_response.status_text,
            raw_body=self.serialize(body),
        )
        resp.body = body
        resp.error_code = body["errorNum"]
        resp.error_message = body["errorMessage"]
        resp.is_success = False
        return resp

    def normalize_data(self, data: Any) -> Union[str, MultipartEncoder, None]:
        """Normalize request data.

        :param data: Request data.
        :type data: str | MultipartEncoder | None
        :return: Normalized data.
        :rtype: str | MultipartEncoder | None
        """
        if data is None:
            return None
        elif isinstance(data, (str, MultipartEncoder)):
            return data
        else:
            return self.serialize(data)

    def ping(self) -> int:
        """Ping the next host to check if connection is established.

        :return: Response status code.
        :rtype: int
        """
        request = Request(method="get", endpoint="/dateToBlock")
        resp = self.send_request(request)
        if resp.status_code in {401, 403}:
            raise ServerConnectionError(f"bad api key ")
        if not resp.is_success:  # pragma: no cover
            raise ServerConnectionError(resp.error_message or "bad server response")
        return resp.status_code

    @abstractmethod
    def send_request(self, request: Request) -> Response:  # pragma: no cover
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        raise NotImplementedError


class BasicConnection(BaseConnection):
    """Connection to specific Moralis using basic authentication.
    :param host
    :type host: str
    :param x_api_keys: list Api key for moralis
    :type x_api_keys: [str]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]\
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    def __init__(
            self,
            host: str,
            x_api_keys: Fields,
            host_resolver: HostResolver,
            sessions: Sequence[Session],
            http_client: HTTPClient,
            serializer: Callable[..., str],
            deserializer: Callable[[str], Any],
    ) -> None:
        super().__init__(
            host,
            x_api_keys,
            host_resolver,
            sessions,
            http_client,
            serializer,
            deserializer,
        )

    def send_request(self, request: Request) -> Response:
        """Send an HTTP request to Moralis server.

        :param request: HTTP request.
        :type request: moralis_services.request.Request
        :return: HTTP response.
        :rtype: moralis_services.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        x_api_key = self._x_api_keys[host_index]
        request.headers["X-API-Key"] = x_api_key

        resp = self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefix + request.endpoint,
            params=request.params,
            data=self.normalize_data(request.data),
            headers=request.headers,
        )
        return self.prep_response(resp, request.deserialize)



from typing import List

import requests
from query_state_lib.base.executors.batch_work_executor import RETRY_EXCEPTIONS

from query_state_lib.base.mappers.eth_call_balance_of_mapper import generate_eth_call_balance_of_json_rpc
from query_state_lib.base.mappers.eth_call_mapper import generate_eth_call_json_rpc
from query_state_lib.base.mappers.eth_json_rpc_mapper import EthJsonRpc, generate_eth_json_rpc
from query_state_lib.base.mappers.get_balance_mapper import generate_get_balance_json_rpc
from query_state_lib.base.providers.auto import get_provider_from_uri
from query_state_lib.jobs.sent_batch_request_job import SentBatchRequestJob


def _sent_batch(list_json_rpc: List[EthJsonRpc], func_sent_handler, err_list=[]):
    """

    :param list_json_rpc:
    :param func_sent_handler:   is a function with argument is a request
                                request is a list , each element is a dict with form of json rpc like :
                                    {
                                        'jsonrpc': '2.0',
                                        'method': method,
                                        'params': params,
                                        'id': request_id,
                                    }

    :return:
    """
    if not list_json_rpc:
        return dict()
    dict_eth_json_rpc = dict()
    type_dict_list = dict()
    for json_rpc in list_json_rpc:
        type_list = type_dict_list.get(json_rpc.type)
        if not type_list:
            type_dict_list[json_rpc.type] = []
            type_list = type_dict_list[json_rpc.type]
        type_list.append(json_rpc)
        dict_eth_json_rpc[json_rpc.id] = json_rpc
    request = []
    for type in type_dict_list:
        request += generate_json_rpc_from_type(type, type_dict_list[type])

    response = func_sent_handler(request)
    check_response(response)
    for response_item in response:
        id = response_item.get("id")
        result = response_item.get("result")
        dict_eth_json_rpc[id].set_result(result)
        if not result:
            err_list.append(dict_eth_json_rpc[id])
            dict_eth_json_rpc[id].error = response_item.get("error")
    return dict_eth_json_rpc


def check_response(response):
    res_start = response[0]
    id = res_start.get("id")
    err = res_start.get("error")
    res_end = response[-1]
    id2 = res_end.get("id")
    err2 = res_end.get("error")
    if res_start.get("error") and res_end.get("error"):
        raise Exception(f"""
        Response data err 
        Request {id} - err {err}
        ...
        Request {id2}- err {err2}
        """)


class ClientQuerier:
    def __init__(self, provider_url):
        self.provider_url = provider_url
        self.batch_provider = get_provider_from_uri(provider_url, batch=True)
        self.recent_err = []

    def sent_batch_to_state_querier(self, list_json_rpc: List[EthJsonRpc]):
        """

        :param list_json_rpc:
        :return:
        """

        url = self.provider_url + "/batch-query"

        def sent_batch_to_state_querier_server(request):
            data = {
                "request": request,
                "batch_request": True
            }
            res = requests.post(url, json=data)
            if res.status_code != 200:
                raise Exception(f"provider {self.provider_url} not support batch-query api")
            return res.json().get("response")

        sent_handler = sent_batch_to_state_querier_server
        self.recent_err = []
        dict_eth_json_rpc = _sent_batch(list_json_rpc, sent_handler, self.recent_err)

        return dict_eth_json_rpc

    def sent_batch_to_provider(self, list_json_rpc: List[EthJsonRpc], batch_size=2000, max_workers=8,
                               sleep_time_retries=10, timeout=120, max_retries=3,
                               retry_exceptions=RETRY_EXCEPTIONS):
        """

        :param list_json_rpc:
        :return:
        """

        batch_provider = self.batch_provider

        def sent_batch_direct(request):
            job = SentBatchRequestJob(request, batch_provider, batch_size=batch_size, max_workers=max_workers,
                                      sleep_time_retries=sleep_time_retries, timeout=timeout, max_retries=max_retries,
                                      retry_exceptions=retry_exceptions, )
            job.run()

            response = job.get_response()
            return response

        sent_handler = sent_batch_direct
        self.recent_err = []
        dict_eth_json_rpc = _sent_batch(list_json_rpc, sent_handler, self.recent_err)

        return dict_eth_json_rpc


MAP_TYPE_JSON_RPC_GENERATOR = {
    "eth_get_balance": generate_get_balance_json_rpc,
    "eth_call_balance_of": generate_eth_call_balance_of_json_rpc,
    "eth_call": generate_eth_call_json_rpc,
    "eth_json_rpc": generate_eth_json_rpc
}


def generate_json_rpc_from_type(type, eth_json_rpc_list: List[EthJsonRpc]):
    generator = MAP_TYPE_JSON_RPC_GENERATOR.get(type)
    if generator:
        return list(generator(eth_json_rpc_list))
    else:
        return []



from typing import List

from query_state_lib.base.mappers.eth_call_mapper import EthCall, generate_eth_call_json_rpc
from query_state_lib.base.utils.decoder import decode_eth_call_balance_of
from query_state_lib.base.utils.encoder import encode_eth_call_balance_of


class EthCallBalanceOf(EthCall):
    def __init__(self, contract_address="", address="", block_number="", id=0):
        data = encode_eth_call_balance_of(address)
        super().__init__(to=contract_address, data=data, block_number=block_number, id=id)
        self.type = "eth_call_balance_of"
        self.address = address

    def set_result(self, result):
        try:
            self.result = decode_eth_call_balance_of(result)
        except Exception as e:
            print(f"""
            EthCallBalanceOf set result err {e}
            EthCallBalanceOf  result : {result}
            EthCallBalanceOf id : {self.id}
            EthCallBalanceOf smart contract : {self.to}
            EthCallBalanceOf block number : {self.block_number}
            EthCallBalanceOf address : {self.address}
            """)
        return self.result

    def decode_result(self):
        return self.result


def generate_eth_call_balance_of_json_rpc(call_infos: List[EthCallBalanceOf]):
    return generate_eth_call_json_rpc(call_infos)



from typing import Any, Dict, List, MutableMapping, Sequence, Union

Json = Dict[str, Any]
Jsons = List[Json]
Params = MutableMapping[str, Union[bool, int, str]]
Headers = MutableMapping[str, str]
Fields = Union[str, Sequence[str]]



from typing import List


class EthJsonRpc:
    def __init__(self, method="", params=None, id=0):
        self.id = id
        self.type = "eth_json_rpc"
        self.method = method
        self.params = params
        self.result = None
        self.decoded = False
        self.error = None

    def set_result(self, result):
        self.result = result

    def decode_result(self):
        return self.result


def generate_eth_json_rpc(call_infos: List[EthJsonRpc]):
    for call_info in call_infos:
        yield generate_json_rpc(
            method=call_info.method,
            params=call_info.params,
            request_id=call_info.id
        )


def generate_json_rpc(method, params, request_id=1):
    return {
        'jsonrpc': '2.0',
        'method': method,
        'params': params,
        'id': request_id,
    }



# The MIT License (MIT)
#
# Copyright (c) 2016 Piper Merriam
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json
import socket

from web3.providers.ipc import IPCProvider
from web3._utils.threads import (
    Timeout,
)
from web3 import Web3
try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


# Mostly copied from web3.py/providers/ipc.py. Supports batch requests.
# Will be removed once batch feature is added to web3.py https://github.com/ethereum/web3.py/issues/832
# Also see this optimization https://github.com/ethereum/web3.py/pull/849
class BatchIPCProvider(Web3.IPCProvider):
    _socket = None

    def make_batch_request(self, text):
        request = text.encode('utf-8')
        with self._lock, self._socket as sock:
            try:
                sock.sendall(request)
            except BrokenPipeError:
                # one extra attempt, then give up
                sock = self._socket.reset()
                sock.sendall(request)

            raw_response = b""
            with Timeout(self.timeout) as timeout:
                while True:
                    try:
                        raw_response += sock.recv(4096)
                    except socket.timeout:
                        timeout.sleep(0)
                        continue
                    if raw_response == b"":
                        timeout.sleep(0)
                    elif has_valid_json_rpc_ending(raw_response):
                        try:
                            response = json.loads(raw_response.decode('utf-8'))
                        except JSONDecodeError:
                            timeout.sleep(0)
                            continue
                        else:
                            return response
                    else:
                        timeout.sleep(0)
                        continue


# A valid JSON RPC response can only end in } or ] http://www.jsonrpc.org/specification
def has_valid_json_rpc_ending(raw_response):
    for valid_ending in [b"}\n", b"]\n"]:
        if raw_response.endswith(valid_ending):
            return True
    else:
        return False



from typing import List

from query_state_lib.base.mappers.eth_json_rpc_mapper import generate_json_rpc, EthJsonRpc
from query_state_lib.base.utils.decoder import decode_eth_call_data


class EthCall(EthJsonRpc):
    def __init__(self, to="", data="", block_number=0, abi=None, fn_name=None, id=0):
        super().__init__()
        self.type = "eth_call"
        self.to = to
        self.data = data
        self.block_number = block_number
        self.id = id
        self.abi = abi
        self.fn_name = fn_name

    def set_abi(self, abi):
        self.abi = abi

    def set_fn_name(self, fn_name):
        self.fn_name = fn_name

    def decode_result(self):
        if not self.abi:
            raise Exception("Have to set abi before decode eth call result")
        if not self.fn_name:
            raise Exception("Have to set function name (fn_name) before decode eth call result")
        if self.decoded:
            return self.result
        self.result = decode_eth_call_data(self.abi, self.fn_name, self.result)
        self.decoded = True
        return self.result


def generate_eth_call_json_rpc(call_infos: List[EthCall]):
    for call_info in call_infos:
        block = call_info.block_number
        yield generate_json_rpc(
            method="eth_call",
            params=[{
                "to": call_info.to,
                "data": call_info.data
            }, hex(block) if isinstance(block, int) else block],
            request_id=call_info.id
        )



from typing import List

from query_state_lib.base.mappers.eth_json_rpc_mapper import generate_json_rpc, EthJsonRpc
from query_state_lib.base.utils.decoder import decode_eth_get_balance


class GetBalance(EthJsonRpc):
    def __init__(self, address="", block_number="latest", id=0):
        super().__init__()
        self.type = "eth_get_balance"
        self.address = address
        self.block_number = block_number
        self.id = id
        self.result = None

    def set_result(self, result):
        if not result:
            self.result = 0
            return result
        self.result = decode_eth_get_balance(result)
        return self.result


def generate_get_balance_json_rpc(get_balance_infos: List[GetBalance]):
    for get_balance in get_balance_infos:
        block = get_balance.block_number
        yield generate_json_rpc(
            method="eth_getBalance",
            params=[
                get_balance.address
                , hex(block) if isinstance(block, int) else block],
            request_id=get_balance.id
        )



LIQUIDATE_EVENT ={
      "anonymous": False,
      "inputs": [
        {
          "indexed": True,
          "internalType": "address",
          "name": "collateralAsset",
          "type": "address"
        },
        {
          "indexed": True,
          "internalType": "address",
          "name": "debtAsset",
          "type": "address"
        },
        {
          "indexed": True,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": False,
          "internalType": "uint256",
          "name": "debtToCover",
          "type": "uint256"
        },
        {
          "indexed": False,
          "internalType": "uint256",
          "name": "liquidatedCollateralAmount",
          "type": "uint256"
        },
        {
          "indexed": False,
          "internalType": "address",
          "name": "liquidator",
          "type": "address"
        },
        {
          "indexed": False,
          "internalType": "bool",
          "name": "receiveAToken",
          "type": "bool"
        }
      ],
      "name": "LiquidationCall",
      "type": "event"
    }


from typing import Any, MutableMapping, Optional

from query_state_lib.client.moralis_client.typings import Headers, Params, Fields


def normalize_headers(headers: Optional[Headers]) -> Headers:
    normalized_headers: Headers = {
        "charset": "utf-8",
        "content-type": "application/json",
    }
    if headers is not None:
        for key, value in headers.items():
            normalized_headers[key.lower()] = value

    return normalized_headers


def normalize_params(params: Optional[Params]) -> MutableMapping[str, Any]:
    normalized_params: MutableMapping[str, Any] = {}

    if params is not None:
        for key, value in params.items():
            if isinstance(value, bool):
                value = int(value)
            if isinstance(value, list):
                normalized_params[key] = value
            else:
                normalized_params[key] = str(value)

    return normalized_params


class Request:
    """HTTP request.

    :param method: HTTP method in lowercase (e.g. "post").
    :type method: str
    :param endpoint: API endpoint.
    :type endpoint: str
    :param headers: Request headers.
    :type headers: dict | None
    :param params: URL parameters.
    :type params: dict | None
    :param data: Request payload.
    :type data: str | bool | int | float | list | dict | None | MultipartEncoder
    :param deserialize: Whether the response body can be deserialized.
    :type deserialize: bool

    :ivar method: HTTP method in lowercase (e.g. "post").
    :vartype method: str
    :ivar endpoint: API endpoint.
    :vartype endpoint: str
    :ivar headers: Request headers.
    :vartype headers: dict | None
    :ivar params: URL (query) parameters.
    :vartype params: dict | None
    :ivar data: Request payload.
    :vartype data: str | bool | int | float | list | dict | None
    :vartype exclusive: str | [str] | None
    :ivar deserialize: Whether the response body can be deserialized.
    :vartype deserialize: bool
    """

    __slots__ = (
        "method",
        "endpoint",
        "headers",
        "params",
        "data",
        "deserialize",
    )

    def __init__(
            self,
            method: str,
            endpoint: str,
            headers: Optional[Headers] = None,
            params: Optional[Params] = None,
            data: Any = None,
            deserialize: bool = True,
    ) -> None:
        self.method = method
        self.endpoint = endpoint
        self.headers: Headers = normalize_headers(headers)
        self.params: MutableMapping[str, Any] = normalize_params(params)
        self.data = data
        self.deserialize = deserialize



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="QueryStateLib",
    version="1.1.2",
    author="LinLin",
    author_email="nguyenthanh2303@gmail.com",
    description="",
    long_description=long_description,
    long_description_content_type="",
    url="https://github.com/thanhnv2303/QueryStateLib",
    project_urls={
        "Bug Tracker": "https://github.com/thanhnv2303/QueryStateLib",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        'web3>=5.19.0',
        'requests<=2.26.0',
        'pbkdf2==1.3',
    ]
)


# MIT License
#
# Copyright (c) 2018 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import json
import logging

from query_state_lib.base.executors.batch_work_executor import BatchWorkExecutor, RETRY_EXCEPTIONS
from query_state_lib.base.jobs.base_job import BaseJob

logger = logging.getLogger(__name__)


class SentBatchRequestJob(BaseJob):
    def __init__(
            self,
            request,
            batch_provider,
            batch_size=2000,
            max_workers=5,
            sleep_time_retries=10,
            timeout=120,
            max_retries=3,
            retry_exceptions=RETRY_EXCEPTIONS,
    ):
        self.request = request
        self.batch_work_executor = BatchWorkExecutor(batch_size, max_workers, retry_exceptions=retry_exceptions,
                                                     max_retries=max_retries, sleep=sleep_time_retries)
        self.batch_provider = batch_provider
        self.sleep_time_retries = sleep_time_retries
        self.response = []
        self.timeout = timeout

    def _start(self):
        pass

    def _export(self):
        self.batch_work_executor.execute(
            self.request,
            self._export_batch,
            total_items=len(self.request)
        )

    def _export_batch(self, batched_request):
        self.response += self.batch_provider.make_batch_request(json.dumps(batched_request), timeout=self.timeout)

    def _end(self):
        self.batch_work_executor.shutdown()

    def get_response(self):
        return self.response

    def clean_cache(self):
        self.response = []



from web3 import Web3

w3 = Web3()


def encode_eth_call_data(abi, fn_name, args):
    contract = w3.eth.contract(abi=abi)

    encode_fnc = contract.encodeABI(fn_name=fn_name, args=args)
    return encode_fnc


def encode_eth_call_balance_of(address):
    prefix = "0x70a08231000000000000000000000000"
    return prefix + str(address).lower()[2:]



ASCII_0 = 0


def clean_user_provided_content(content):
    if isinstance(content, str):
        # This prevents this error in BigQuery
        # Error while reading data, error message: Error detected while parsing row starting at position: 9999.
        # Error: Bad character (ASCII 0) encountered.
        return content.translate({ASCII_0: None})
    else:
        return content


from typing import Optional

from query_state_lib.client.moralis_client.moralis_request import Request
from query_state_lib.client.moralis_client.moralis_response import Response


class MoralisError(Exception):
    """Base class for all exceptions in python-arango."""


class MoralisClientError(MoralisError):
    """Base class for errors originating from python-arango client.

    :param msg: Error message.
    :type msg: str

    :cvar source: Source of the error (always set to "client").
    :vartype source: str
    :ivar message: Error message.
    :vartype message: str
    """

    source = "client"

    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self.message = msg
        self.error_message = None
        self.error_code = None
        self.url = None
        self.response = None
        self.request = None
        self.http_method = None
        self.http_code = None
        self.http_headers = None


class MoralisServerError(MoralisError):
    """Base class for errors originating from ArangoDB server.

    :param resp: HTTP response.
    :type resp: arango.response.Response
    :param msg: Error message override.
    :type msg: str

    :cvar source: Source of the error (always set to "server").
    :vartype source: str
    :ivar message: Exception message.
    :vartype message: str
    :ivar url: API URL.
    :vartype url: str
    :ivar response: HTTP response object.
    :vartype response: arango.response.Response
    :ivar request: HTTP request object.
    :vartype request: arango.request.Request
    :ivar http_method: HTTP method in lowercase (e.g. "post").
    :vartype http_method: str
    :ivar http_code: HTTP status code.
    :vartype http_code: int
    :ivar http_headers: Response headers.
    :vartype http_headers: dict
    :ivar error_code: Error code from ArangoDB server.
    :vartype error_code: int
    :ivar error_message: Raw error message from ArangoDB server.
    :vartype error_message: str
    """

    source = "server"

    def __init__(
            self, resp: Response, request: Request, msg: Optional[str] = None
    ) -> None:
        msg = msg or resp.error_message or resp.status_text
        self.error_message = resp.error_message
        self.error_code = resp.error_code
        if self.error_code is not None:
            msg = f"[HTTP {resp.status_code}][ERR {self.error_code}] {msg}"
        else:
            msg = f"[HTTP {resp.status_code}] {msg}"
            self.error_code = resp.status_code
        super().__init__(msg)
        self.message = msg
        self.url = resp.url
        self.response = resp
        self.request = request
        self.http_method = resp.method
        self.http_code = resp.status_code
        self.http_headers = resp.headers


#####################
# Server Exceptions #
#####################


class ServerConnectionError(MoralisClientError):
    """Failed to connect to ArangoDB server."""


class ServerEngineError(MoralisServerError):
    """Failed to retrieve database engine."""


class ServerVersionError(MoralisServerError):
    """Failed to retrieve server version."""


class ServerDetailsError(MoralisServerError):
    """Failed to retrieve server details."""


class ServerStatusError(MoralisServerError):
    """Failed to retrieve server status."""


class ServerTimeError(MoralisServerError):
    """Failed to retrieve server system time."""


class ServerEchoError(MoralisServerError):
    """Failed to retrieve details on last request."""


class ServerShutdownError(MoralisServerError):
    """Failed to initiate shutdown sequence."""


class ServerRunTestsError(MoralisServerError):
    """Failed to execute server tests."""


class ServerRequiredDBVersionError(MoralisServerError):
    """Failed to retrieve server target version."""


class ServerReadLogError(MoralisServerError):
    """Failed to retrieve global log."""


class ServerLogLevelError(MoralisServerError):
    """Failed to retrieve server log levels."""


class ServerLogLevelSetError(MoralisServerError):
    """Failed to set server log levels."""


class ServerReloadRoutingError(MoralisServerError):
    """Failed to reload routing details."""


class ServerStatisticsError(MoralisServerError):
    """Failed to retrieve server statistics."""


class ServerMetricsError(MoralisServerError):
    """Failed to retrieve server metrics."""


class ServerRoleError(MoralisServerError):
    """Failed to retrieve server role in a cluster."""


class ServerTLSError(MoralisServerError):
    """Failed to retrieve TLS data."""


class ServerTLSReloadError(MoralisServerError):
    """Failed to reload TLS."""


class ServerEncryptionError(MoralisServerError):
    """Failed to reload user-defined encryption keys."""
